"""Main part of audiness."""
import typer
from typing_extensions import Annotated
from validators import url

from audiness.commands import folders, policies, scans, server, software
from audiness.helpers import setup_connection

app = typer.Typer()

app.add_typer(folders.app, name="folders")
app.add_typer(policies.app, name="policies")
app.add_typer(scans.app, name="scans")
app.add_typer(server.app, name="server")
app.add_typer(software.app, name="software")


def validate_host(value: str):
    """Check if host is a valid URL."""
    if url(value):
        raise typer.BadParameter("URL is not valid")
    return value


def validate_key(value: str):
    """Check the length of an API key."""
    if len(value) != 64:
        raise typer.BadParameter("Key doesn't have the right length")
    return value


@app.callback()
def main(
    ctx: typer.Context,
    access_key: Annotated[
        str,
        typer.Option(
            envvar="ACCESS_KEY",
            help="Nessus API access key",
            prompt=True,
            callback=validate_key,
        ),
    ],
    secret_key: Annotated[
        str,
        typer.Option(
            envvar="SECRET_KEY",
            help="Nessus API secret key",
            prompt=True,
            callback=validate_key,
        ),
    ],
    host: Annotated[
        str, typer.Option(help="URL to Nessus instance", callback=validate_host)
    ] = "https://localhost:8834",
):
    connection = setup_connection(host, access_key, secret_key)
    ctx.obj = {"connection": connection}


if __name__ == "__main__":
    app()



"""Interact with policies in Nessus."""

import typer
from rich.console import Console
from rich.table import Table
from tenable.nessus import Nessus

from audiness.helpers import human_readable_datetime

app = typer.Typer()


@app.command()
def list(ctx: typer.Context):
    """List all policies."""
    connection = ctx.obj.get("connection")

    policies = connection.policies.list()

    human_readable_datetime(policies, "last_modification_date", "creation_date")

    table = Table(title="All available policies")

    table.add_column("Name", justify="left", style="cyan", no_wrap=True)
    table.add_column("Description")
    table.add_column("Creation date")
    table.add_column("Last modification date")

    for policy in policies:
        table.add_row(
            policy["name"],
            policy["description"],
            policy["creation_date"],
            policy["last_modification_date"],
        )

    console = Console()
    console.print(table)



"""Interact with folders in Nessus."""
import typer
from restfly.errors import RequestConflictError
from rich.console import Console
from rich.table import Table
from tenable.nessus import Nessus
from typing_extensions import Annotated

app = typer.Typer()


@app.command()
def list(ctx: typer.Context):
    """List all folders."""
    connection = ctx.obj.get("connection")

    folders = connection.folders.list()

    table = Table(title="All available folders")

    table.add_column("Name", justify="left", style="cyan", no_wrap=True)
    table.add_column("ID")
    table.add_column("Unread Count", justify="right")

    for folder in folders:
        table.add_row(folder["name"], str(folder["id"]), str(folder["unread_count"]))

    console = Console()
    console.print(table)


@app.command()
def create(
    ctx: typer.Context,
    folder_name: Annotated[
        str,
        typer.Option(help="Name of the folder to create", prompt=True),
    ],
):
    """Create a new folder."""
    connection = ctx.obj.get("connection")

    try:
        connection.folders.create(folder_name)
    except RequestConflictError:
        return

    for folder in connection.folders.list():
        if folder["name"] == folder_name:
            print(
                f'Folder {folder_name} has ID {folder["id"]} and was successfully created.'
            )



"""Interact with the Nessus instance."""

import typer
from rich.console import Console
from rich.progress import track
from rich.table import Table

from audiness.helpers import human_readable_datetime

app = typer.Typer()


@app.command()
def status(ctx: typer.Context):
    """Get the status of the Nessus instance."""
    connection = ctx.obj.get("connection")

    status = connection.server.status()

    print(status)


@app.command()
def alerts(ctx: typer.Context):
    """Get the alerts of the Nessus instance."""
    connection = ctx.obj.get("connection")

    alerts = connection.settings.alerts()

    table = Table(title="Alerts")

    table.add_column("Alert", justify="left", style="cyan", no_wrap=True)
    table.add_column("Description")
    table.add_column("Type")
    table.add_column("Severity")

    for alert in alerts:
        table.add_row(
            alert["alert"],
            alert["description"],
            alert["type"],
            str(alert["severity"]),
        )

    console = Console()
    console.print(table)


@app.command()
def status(ctx: typer.Context):
    """Get the status of the Nessus instance."""
    connection = ctx.obj.get("connection")

    status = connection.settings.health()["perf_stats_history"]

    human_readable_datetime(status, "timestamp")

    table = Table(title="Status")

    table.add_column("Timestamp", justify="left", style="cyan", no_wrap=True)
    table.add_column("kBytes received")
    table.add_column("kBytes sent")
    table.add_column("Nessus RAM")

    for entry in status:
        table.add_row(
            entry["timestamp"],
            str(entry["kbytes_received"]),
            str(entry["kbytes_sent"]),
            str(entry["nessus_mem"]),
        )

    console = Console()
    console.print(table)



"""Interact with scans in Nessus."""

import io
import time
from datetime import datetime
from pathlib import Path
from typing import Optional

import typer
from rich.console import Console
from rich.progress import track
from rich.table import Table
from typing_extensions import Annotated

from audiness.helpers import human_readable_datetime

app = typer.Typer()


@app.command()
def list(ctx: typer.Context):
    """List all available scans."""
    connection = ctx.obj.get("connection")
    scans = connection.scans.list()["scans"]

    human_readable_datetime(scans, "creation_date", "last_modification_date")

    table = Table(title="All available scans")

    table.add_column("Name", justify="left", style="cyan", no_wrap=True)
    table.add_column("ID")
    table.add_column("Last modification date")
    table.add_column("Status")
    table.add_column("Folder ID")

    for scan in scans:
        table.add_row(
            scan["name"],
            str(scan["id"]),
            scan["last_modification_date"],
            scan["status"],
            str(scan["folder_id"]),
        )

    console = Console()
    console.print(table)


@app.command()
def export(
    ctx: typer.Context,
    identifier: Annotated[
        str, typer.Option(help="String for the identification", prompt=True)
    ] = "SAS",
    path: Annotated[
        Optional[Path],
        typer.Option(help="Path to store the exported files", prompt=True),
    ] = ".",
    history: Annotated[
        int,
        typer.Option(help="Historical state to export", prompt=True),
    ] = 1,
):
    """Export a Nessus scan or Nessus scans."""
    connection = ctx.obj.get("connection")

    # Get all scans
    scans = connection.scans.list()

    # Clean scans result and limit data to what's needed
    del scans["timestamp"]
    del scans["folders"]

    failed_exports = []
    relevant_scans = []

    for scan in scans["scans"]:
        if scan["name"].startswith(identifier) and scan["status"] == "completed":
            relevant_scans.append(scan)

    # Export all scans which matches the identifier
    for scan in track(relevant_scans, description="Processing scans ..."):
        scan_history = connection.scans.details(scan["id"])["history"]

        try:
            single_scan = scan_history[-abs(history)]

            short_date = datetime.fromtimestamp(
                single_scan["last_modification_date"]
            ).strftime("%Y%m")
            historic_scan_id = single_scan["history_id"]

            filename = Path(path) / Path(f'{scan["name"]}_{short_date}.nessus')

            scan_data = connection.scans.export_scan(
                scan_id=scan["id"], history_id=historic_scan_id, format="nessus"
            )
            Path(filename).write_bytes(scan_data.getbuffer())
        except IndexError:
            failed_exports.append(scan["name"])

    print(
        f"{len(relevant_scans)} files selected, {len(relevant_scans)-len(failed_exports)} exported, {len(failed_exports)} failed."
    )
    if failed_exports:
        print(f"Failed exports:")
        for element in failed_exports:
            print(element)



"""Helper elements for various tasks."""
from datetime import datetime

from tenable.nessus import Nessus


def setup_connection(host, access_key, secret_key):
    """Connect to the Nessus instance."""
    connection = Nessus(url=host, access_key=access_key, secret_key=secret_key)
    return connection


def human_readable_datetime(data, *args):
    """Replace timestamps with a human readable date in a dict."""
    format_string = "%Y-%m-%d %H:%M:%S"
    for entry in data:
        for element in args:
            entry[element] = datetime.fromtimestamp(entry[element]).strftime(
                format_string
            )

    return data



"""Trigger an update of Nessus."""
import typer

app = typer.Typer()


@app.command()
def update(ctx: typer.Context):
    """Update Nessus."""
    connection = ctx.obj.get("connection")

    connection.software_update.update()    
